// LICENSE_CODE ZON
'use strict'; /*jslint browser:true, es5:true, devel:true*/ /*global exports*/
var sdk_self = require('sdk/self');
var data_url = sdk_self.data.url;
var tabs = require('sdk/tabs');
var tabutil = require('sdk/tabs/utils');
var tabworker = require('sdk/tabs/worker');
var winutil = require('sdk/window/utils');
var storage = require('sdk/simple-storage').storage;
/* private-browsing/window/utils undocumented. not available in FF32 */
var pbwinutil = null;
try { pbwinutil = require('sdk/private-browsing/window/utils'); }
catch(e){}
var sdk_windows = require('sdk/windows');
var windows = sdk_windows.browserWindows;
var BrowserWindow = sdk_windows.BrowserWindow;
var is_private = require('sdk/private-browsing').isPrivate;
var EventTarget = require('sdk/event/target').EventTarget;
var events = require('sdk/event/core');
var emit = events.emit;
var pagemod = require('sdk/page-mod');
var symbiont = null;
try { symbiont = require('sdk/content/symbiont'); }
catch(e){ symbiont = require('sdk/content/content'); }
var hiddenframes = require('sdk/frame/hidden-frame');
var panel = require('sdk/panel').Panel;
var uuid = require('sdk/util/uuid').uuid;
var timers = require('sdk/timers');
var system = require('sdk/system');
var sysevents = require('sdk/system/events');
var model, view; // since FF31
try {
    model = require('sdk/model/core');
    view = require('sdk/view/core');
} catch(e){}
var chrome = require('chrome');
var Cc = chrome.Cc;
var Ci = chrome.Ci;
var Cu = chrome.Cu;
var baction = require('browserAction').BrowserAction;
var _ = require('./underscore');
var _conf = require('./conf');
var conf = _conf.conf;
var zconf = _conf.zon_config;
var array = require('./array');
var proxy = require('./proxy');
var plugin = require('./plugin');
var util = require('./ff_util');
var QI = util.QI;
var version_util = require('./version_util');
var E = _.extend(exports, {storage: {}, tabs: {}, impl: {}, windows: {},
    runtime: {}, be: {}, firefox: {}, extension: {}, proxy: {},
    management: {}, cookies: {}, webRequest: {}, webNavigation: {}});
var services = Cu.import('resource://gre/modules/Services.jsm', {}).Services;
var prompts = services.prompt, io = services.io, cookies = services.cookies;
var AddonManager = Cu.import('resource://gre/modules/AddonManager.jsm', {})
.AddonManager;
var downloader = Cc['@mozilla.org/network/downloader;1'];
var fileutil = Cu.import('resource://gre/modules/FileUtils.jsm', {}).FileUtils;
var netutil = Cu.import('resource://gre/modules/NetUtil.jsm', {}).NetUtil;
var registry = Cc['@mozilla.org/windows-registry-key;1'];
var br = util.gBrowser();

function on_open(tab){ emit(E.tabs.onCreated, 'event', to_chrome_tab(tab)); }

function on_close(tab){
    tab = to_chrome_tab(tab);
    emit(E.tabs.onRemoved, 'event', tab.id, {windowId: tab.windowId});
}

function on_activate(tab){
    tab = to_chrome_tab(tab);
    emit(E.tabs.onActivated, 'event',
	{tabId: tab.id, windowId: tab.windowId});
}

function on_ready(tab){
    /* XXX bahaa: test it fires for cached sites (or use tab.on('pageshow')) */
    tab = to_chrome_tab(tab);
    emit(E.tabs.onUpdated, 'event', tab.id, {url: tab.url}, tab);
}

function on_error(event){
    var tab = br.getBrowserIndexForDocument(event.originalTarget);
    if (tab==-1)
        return;
    tab = tabs[tab];
    var win = event.originalTarget.defaultView;
    var webnav = win.QueryInterface(Ci.nsIInterfaceRequestor)
        .getInterface(Ci.nsIWebNavigation);
    var docuri = webnav.document.documentURI;
    var location = win.location+'';
    if (win.frameElement)
      return;
    if (docuri.indexOf('about:neterror')===0)
    {
        tab = to_chrome_tab(tab);
        emit(E.webNavigation.onErrorOccurred, 'event', {tabId: tab && tab.id,
            frameId: 0});
    }
}

function on_headers_received(data){
    emit(E.webRequest.onHeadersReceived, 'event', data);
}

function windows_on_activated(win){
    emit(E.windows.onFocusChanged, 'event', get_win_id(win)); }

function tooltip_fix(panel){
    if (!panel)
        return;
    try { panel.setAttribute('tooltip', 'aHTMLTooltip'); }
    catch(err){ console.error('tooltip_fix: '+err); }
}

function on_popupshowing(e){
    try {
        var p = e.subject, a = p && p.anchorNode, id = a && a.id;
        // if it's anchored on our widget then it's our panel
        if (!id || !id.startsWith('widget:'+sdk_self.id))
            return;
        tooltip_fix(p);
        sysevents.off('popupshowing', on_popupshowing);
    } catch(err){ console.error('on_popupshowing: '+err); }
}

function setup_tooltip_fix(popup){
    if (view && view.getActiveView)
        return void tooltip_fix(view.getActiveView(popup));
    sysevents.on('popupshowing', on_popupshowing);
}

var inited, uninstall_url, install_details;
E.init = function(_install_details){
    if (inited)
	return;
    inited = true;
    install_details = _install_details;
    tabs.on('open', on_open);
    tabs.on('close', on_close);
    tabs.on('activate', on_activate);
    tabs.on('ready', on_ready);
    br.addEventListener('DOMContentLoaded', on_error, false);
    windows.on('activate', windows_on_activated);
    plugin.on('event', on_plugin_event);
    E.be.ccgi.init();
    pagemod.PageMod({
        include: data_url('be_about.html')+'*',
        contentScriptFile: data_url('content_script.js'),
        contentScriptWhen: 'start',
        contentScriptOptions: {type: 'about'},
        onAttach: function(w){
            new Worker('about', w);
            w.on('message', function(m){ on_msg('about', m); });
        },
        attachTo: ['existing', 'top'],
    });
};
E.uninit = function(){
    if (!inited)
	return;
    tabs.removeListener('open', on_open);
    tabs.removeListener('close', on_close);
    tabs.removeListener('activate', on_activate);
    tabs.removeListener('ready', on_ready);
    br.removeEventListener('DOMContentLoaded', on_error, false);
    windows.removeListener('activate', windows_on_activated);
    sysevents.off('popupshowing', on_popupshowing);
    plugin.removeListener('event', on_plugin_event);
    workers_uninit();
    E.be.ccgi.uninit();
    E.tabs.uninit();
    uninstall_url = install_details = undefined;
    inited = false;
};

E.on_unload = function(reason){
    var url = uninstall_url;
    console.log('api.on_unload '+reason+' '+url);
    E.uninit();
    if (!url)
	return;
    /* XXX arik/bahaa: we will never recieve uninstall because firefox has a
     * bug where it sends 'disable' event for uninstall, see comment in:
     * https://addons.mozilla.org/en-US/developers/docs/sdk/latest/dev-guide/
     * tutorials/load-and-unload.html */
    if (['uninstall', 'disable'].indexOf(reason)==-1)
	return;
    try {
	tabs.open({url: url});
	storage.be_disabled = 1;
    } catch(err){ console.error('error in unload '+(err.stack||err)); }
};

function to_chrome_tab(tab){
    var active = active_tab();
    return !tab ? tab : {
        id: tab.id,
        index: tab.index,
        windowId: get_win_id(tab.window), /* tab.window undocumented */
        openerTabId: null,
        highlighted: null,
        active: active && active.id===tab.id,
        pinned: tab.isPinned,
        url: tab.url,
        title: tab.title,
        favIconUrl: null, /* tab.favicon is a base64 encoded png not url */
        status: tab_status(tab),
        incognito: is_private(tab),
        width: null,
        height: null,
    };
}

function active_tab(){
    var win = winutil.getMostRecentBrowserWindow();
    var xtab = win && tabutil.getSelectedTab(win);
    return xtab && from_XUL_tab(xtab);
}

function to_chrome_win(win){ return !win ? win : {id: get_win_id(win)}; }

function to_XUL_tab(tab){
    if (view)
        return view.viewFor(tab);
    return _.find(tabutil.getTabs(), function(t){
        return tabutil.getTabId(t)===tab.id; });
}

function from_XUL_tab(tab){
    if (model)
        return model.modelFor(tab);
    return get_tab(tabutil.getTabId(tab));
}

function to_DOM_window(win){
    if (view)
        return view.viewFor(win);
    if (pbwinutil)
        return pbwinutil.getOwnerWindow(win);
    var tab = win.tabs && win.tabs[0];
    var xtab = tab && to_XUL_tab(tab);
    return xtab && tabutil.getOwnerWindow(xtab);
}

function from_DOM_window(win){
    if (model)
        return model.modelFor(win);
    return new BrowserWindow(win);
}

function no_tab(ids){ return 'no tab with id: '+([].concat(ids)).join(','); }

function get_tab(id){ return _.findWhere(tabs, {id: id}); }

function get_win_id(win){ return get_DOM_win_id(to_DOM_window(win)); }

function get_DOM_win_id(win){ return win ? winutil.getOuterId(win) : -1; }

function tab_status(tab){
    if (1) // XXX bahaa: enable and implement change event
        return null;
    var xtab = tab && to_XUL_tab(tab);
    var br = xtab && tabutil.getBrowserForTab(xtab);
    var wp = br && br.webProgress;
    return !wp ? null : wp.isLoadingDocument ? 'loading' : 'complete';
}

E.impl.init = function(cb){
    cb(null, {id: sdk_self.id,
	url: data_url('').replace(/\/data\/$/, ''), manifest: {}});
};

E.storage.sync = E.storage.local = {};
E.storage.local.get = function(keys, cb){
    try {
        var k, obj, ret = {};
        if (typeof keys==='object')
        {
            k = Object.keys(keys);
            obj = true;
        }
        else
            k = [].concat(keys);
        for (var i=0, l=k.length; i<l; i++)
        {
            ret[k[i]] = storage.hasOwnProperty(k[i]) ?
                storage[k[i]] : obj ? keys[k[i]] : undefined;
        }
        cb(null, ret);
    } catch(e){ cb(''+e); }
};
E.storage.local.set = function(items, cb){
    try {
        for (var key in items)
            storage[key] = items[key];
        cb();
    } catch(e){ cb(''+e); }
};
E.storage.local.remove = function(keys, cb){
    try {
        keys = [].concat(keys);
        for (var i=0, l=keys.length; i<l; i++)
            delete storage[keys[i]];
        cb();
    } catch(e){ cb(''+e); }
};

E.tabs.workers = {};
E.tabs.create = function(details, cb){
    try {
        tabs.open({
            url: details.url,
            onOpen: function(tab){
                cb(null, to_chrome_tab(tab));
            },
            isPrivate: is_private(active_tab()),
        });
    } catch(e){ cb(''+e); }
};
E.tabs.reload = function(id, details, cb){
    try {
        var t = get_tab(id);
        if (t)
            t.reload();
        cb(t ? null : no_tab(id));
    } catch(e){ cb(''+e); }
};
E.tabs.update = function(id, details, cb){
    try {
        var t = get_tab(id);
        if (!t)
            return cb(no_tab(id));
        if (details.url)
            t.url = details.url;
        if (details.active)
            t.activate();
        if (details.pinned===true)
            t.pin();
        else if (details.pinned===false)
            t.unpin();
        return cb(null, to_chrome_tab(t));
    } catch(e){ cb(''+e); }
};
E.tabs.get = function(id, cb){
    try {
        var t = get_tab(id);
        cb(t ? null : no_tab(id), to_chrome_tab(t));
    } catch(e){ cb(''+e); }
};
E.tabs.query = function(info, cb){
    try {
        var ret = [], q = {}, activewin = windows.activeWindow, re;
        var ts = info.lastFocusedWindow===true ? activewin.tabs : tabs;
        try { re = info.url_re && new RegExp(info.url_re);
        } catch(e){ return cb(''+e, []); }
        for (var i=0, l=ts.length; i<l; i++)
        {
            var t = to_chrome_tab(ts[i]);
            if (info.active!==undefined && info.active!==t.active)
                continue;
            if (info.pinned!==undefined && info.pinned!==t.pinned)
                continue;
            if (info.lastFocusedWindow===false && ts[i].window===activewin)
                continue;
            if (info.title!==undefined && info.title!==t.title)
                continue;
            if (re && !re.test(t.url))
                continue;
            if (info.windowId!==undefined && info.windowId!==t.windowId)
                continue;
            if (info.index!==undefined && info.index!==t.index)
                continue;
            ret.push(t);
        }
        cb(null, ret);
    } catch(e){ cb(''+e, []); }
};
E.tabs.data_url = function(file, cb){ return cb(null, data_url(file)); };
E.tabs.executeScript = function(id, details, cb){
    try {
        var t = get_tab(id), cid = details.connection_id;
        if (!t)
            return cb(no_tab(id));
        var opt = {};
        if (details.code)
            opt.contentScript = details.code;
        if (details.file)
            opt.contentScriptFile = _.map([].concat(details.file), data_url);
        if (details.tpopup)
        {
            if (!cid)
                return cb('tpopup must have connection id');
            bg_connection_update(cid, id, true);
        }
        var w = t.attach(opt), ws = [w];
        if (details.allFrames)
        {
            var xtab = to_XUL_tab(t);
            var win = xtab && tabutil.getTabContentWindow(xtab);
            _.each(win && winutil.getFrames(win), function(f){
                ws.push(tabworker.Worker(opt, f)); });
        }
	if (details.tpopup)
        {
            new Worker(cid, w, {tpopup: 1, tab_id: id});
            w.on('message', on_msg.bind(null, cid));
        }
        else
        {
            _.each(ws, function(_w){
                if (details.ccgi)
                    E.be.ccgi.on_attach(_w);
                E.tabs.on_attach(id, _w, _w===w);
            });
        }
        cb(null);
    } catch(e){ cb(''+e); }
};
E.tabs.insertCSS = function(id, details, cb){
    return E.tabs.executeScript(id, details, cb); };
E.tabs.remove = function(ids, cb){
    try {
        var notfound = [];
        _.each([].concat(ids), function(id){
            var t = get_tab(id);
            if (t)
                return t.close();
            notfound.push(id);
        });
        cb(notfound.length ? no_tab(notfound) : null);
    } catch(e){ cb(''+e); }
};
E.tabs.disconnect = function(id, cb){
    _.invoke(_.where(workers, {tab_id: id}), 'destroy'); };
E.tabs.sendMessage = function(id, msg, opt, cb){
    opt = opt||{};
    if (!cb)
    {
        cb = opt;
        opt = {};
    }
    var ws = E.tabs.workers[id];
    _.each(opt.frameId ? _.pick(ws, opt.frameId) : ws, function(w, frame_id){
        try { w.postMessage(msg); }
        catch(e){ console.error('failed post to frame '+frame_id+': '+e); }
    });
    cb(null);
};
E.tabs.on_attach = function(tab_id, w, is_main){
    var t = E.tabs.workers[tab_id] = E.tabs.workers[tab_id]||{};
    var frame_id = is_main ? tab_id+'.0' : _.uniqueId(tab_id+'.');
    t[frame_id] = w;
    w.on('detach', function(){
        delete t[frame_id];
        if (_.isEmpty(t))
            t = void delete E.tabs.workers[tab_id];
    });
    w.port.on('message', function(msg){
        emit(E.tabs.on_message, 'event', msg, {
            tab: to_chrome_tab(get_tab(tab_id)),
            frameId: frame_id,
            url: null, // XXX bahaa: url of iframe
        });
    });
};
E.tabs.uninit = function(){
    _.each(E.tabs.workers, function(l){ _.invoke(l, 'destroy'); });
    E.tabs.workers = {};
};
E.tabs.onCreated = new EventTarget();
E.tabs.onUpdated = new EventTarget();
E.tabs.onActivated = new EventTarget();
E.tabs.onRemoved = new EventTarget();
E.tabs.on_replaced = new EventTarget();
E.tabs.on_message = new EventTarget();

E.webRequest.onHeadersReceived = new EventTarget();
E.webNavigation.onErrorOccurred = new EventTarget();

E.windows.getLastFocused = function(details, cb){
    try { cb(null, to_chrome_win(windows.activeWindow)); }
    catch(e){ cb(''+e); }
};
E.windows.onFocusChanged = new EventTarget();

E.runtime.setUninstallUrl = function(url, cb){
    uninstall_url = url;
    cb();
};
var checking_update, installing;
E.runtime.requestUpdateCheck = function(cb){
    if (checking_update)
        return cb(null, 'throttled');
    checking_update = true;
    var _cb = function(){
        checking_update = false;
        cb.apply(null, arguments);
    };
    var ret = null;
    AddonManager.getAddonByID(sdk_self.id, function(addon){
        if (!addon)
            _cb('this addon was not found?!');
        addon.findUpdates({
            onUpdateAvailable: function(addon, install){
                ret = 'update_available';
                var v = install.version;
                if (installing)
                {
                    return console.error('installation of v'+v+
                    ' already in progress');
                }
                console.log('new version available v'+v);
                install.addListener({
                    onDownloadStarted: function(){
                        console.log('download started'); },
                    onDownloadEnded: function(){
                        console.log('download ended'); },
                    onDownloadFailed: function(){
                        installing = false;
                        console.error('download failed');
                        util.perr('upgrade_err', 'download', ''+install.error);
                    },
                    onInstallStarted: function(){
                        console.log('install started'); },
                    onInstallEnded: function(){
                        installing = false;
                        console.log('upgraded to v'+v);
                    },
                    onInstallFailed: function(){
                        installing = false;
                        console.error('install failed');
                        util.perr('upgrade_err', 'install', ''+install.error);
                    },
                });
                installing = true;
                install.install();
            },
            onNoUpdateAvailable: function(){ ret = 'no_update'; },
            onUpdateFinished: function(addon, err){
                _cb(err ? 'update check failed: '+err : null, ret);
            },
        }, AddonManager.UPDATE_WHEN_PERIODIC_UPDATE);
    });
};
E.runtime.get_install_details = function(cb){ cb(null, install_details); };

var ba = null;
E.browserAction = {};
E.browserAction.init = function(opt){
    E.browserAction.destroy();
    ba = baction(opt);
};
E.browserAction.destroy = function(){
    if (!ba)
        return;
    try { ba.destroy(); }
    catch(e){ console.error('browserAction.destroy()', e); }
    ba = null;
};
_.each(['setTitle', 'setIcon', 'setBadgeText', 'setBadgeBackgroundColor'],
    function(name){
    E.browserAction[name] = function(){
        var args = array.args(arguments), cb = args.pop();
        if (!ba)
            return cb('browserAction not inited');
        try {
            ba[name].apply(ba, args);
            cb();
        } catch(e){ cb(''+e); }
    };
});
_.each(['enable', 'disable', 'setPopup'], function(name){
    E.browserAction[name] = function(){
        var args = array.args(arguments), cb = args.pop();
        cb('unimplemented');
    };
});

E.extension.isAllowedIncognitoAccess = function(cb){ cb(null, true); };

E.proxy.settings = {};
E.proxy.settings.set = function(opt, cb){
    var url, pac;
    if (opt && opt.value && (pac=opt.value.pacScript))
    {
	if (pac.url)
	    url = pac.url;
	else if (pac.data)
	    url = 'data:text/javascript,'+pac.data;
    }
    proxy.set(url);
    cb();
};
E.proxy.settings.clear = function(opt, cb){
    proxy.set('');
    cb();
};

E.be.proxy = {};
E.be.proxy.add_tab = function(tabid, details, cb){
    // XXX bahaa: details no longer used
    proxy.add_tab(tabid);
    cb();
};
E.be.proxy.del_tab = function(tabid, cb){
    proxy.del_tab(tabid);
    cb();
};

function call_bg_func(name/*, arg0, arg1, ... */){
    var args = array.args(arguments).slice(1);
    if (name=='on_headers_received')
        on_headers_received.apply(null, args);
    if (!workers.bg)
        return void console.error('no background');
    var win = workers.bg.window();
    if (!win)
        return void console.error('call_bg_func no window');
    if (!win[name])
        return void console.error('call_bg_func no function '+name);
    // clone needed for new FF versions
    // https://bugzilla.mozilla.org/show_bug.cgi?id=1107882#c8
    if (Cu.cloneInto)
        args = Cu.cloneInto(args, win);
    return win && win[name] && win[name].apply(null, args);
}

_.each(['on_before_request', 'on_before_send_headers', 'on_headers_received',
    'on_completed', 'on_find_proxy_for_url', 'on_auth_required'],
    function(obj){
    var o = E.be.proxy[obj] = {};
    o.set_listener = function(name, cb){
        proxy[obj].set_listener(call_bg_func.bind(null, name));
        cb();
    };
    o.clear_listener = function(cb){
        proxy[obj].clear_listener();
        cb();
    };
});

// XXX arik/bahaa: try to find firefox native api for reload
E.be.reload_ext = function(cb){
    // XXX arik: support reload when there are debug tabs
    console.log('be.reload_ext');
    install_details = undefined;
    E.be.plugin.stop();
    E.browserAction.destroy();
    workers_uninit();
    E.tabs.uninit();
    E.be.ccgi.uninit();
    E.be.ccgi.init();
    E.create_background();
    E.create_popup();
};

var ccgi = E.be.ccgi = new EventTarget();
ccgi.worker_id = 0;
ccgi.on_attach = function(w){
    var id = ccgi.worker_id++;
    ccgi.workers[id] = w;
    w.on('detach', function(){ delete ccgi.workers[id]; });
    w.port.on('req', function(msg){
        emit(ccgi, 'event', msg, id, {tab: to_chrome_tab(w.tab)});
    });
};
ccgi.init = function(){
    if (ccgi.inited)
        return;
    ccgi.inited = true;
    ccgi.workers = {};
    ccgi.pagemods = [];
    ccgi.pagemods.push(pagemod.PageMod({
        include: new RegExp(conf.hola_re),
        contentScriptFile: data_url('cs_hola.js'),
        contentScriptWhen: 'start',
        contentScriptOptions:
            {ff_debug_page: conf.url_bext_debug+'/be_test.html'},
        onAttach: ccgi.on_attach,
        attachTo: ['existing', 'top'],
    }));
    // XXX colin: contentScriptWhen should be 'start', doesn't attach onto
    // iframes with src="" or src="about:blank"
    ccgi.pagemods.push(pagemod.PageMod({
        include: ['*'],
        contentScriptFile: [data_url('jquery.min.js'),
            data_url('cs_inject.js')],
        contentScriptWhen: 'end',
        onAttach: function(w){
            ccgi.on_attach(w);
            E.tabs.on_attach(w.tab.id, w, false);
        },
        attachTo: ['existing', 'top', 'frame'],
    }));
};
ccgi.send = function(msg, id, cb){
    var w = ccgi.workers[id], e = null;
    if (!w)
        return cb('no ccgi worker with id: '+id);
    try { w.port.emit('resp', msg); }
    catch(ex){ e = ''+ex; }
    cb(e);
};
ccgi.uninit = function(){
    if (!ccgi.inited)
        return;
    ccgi.inited = false;
    _.invoke(ccgi.pagemods, 'destroy');
    _.invoke(ccgi.workers, 'destroy');
    ccgi.workers = null;
};

E.be.plugin = new EventTarget();
E.be.plugin.init = function(cb){ plugin.init(cb); };
E.be.plugin.start = function(cb){ plugin.start(cb); };
E.be.plugin.stop = function(cb){ plugin.stop(cb); };
E.be.plugin.is_running = function(cb){ cb(null, plugin.is_running()); };
E.be.plugin.get_info = function(cb){ cb(null, plugin.get_info()); };
E.be.plugin.perr = function(opt, cb){
    plugin.perr(opt);
    cb();
};
E.be.plugin.add_fw_rule = function(path, cb){
    plugin.add_fw_rule(path, cb); };
E.be.plugin.exec = function(exe, args, cb){
    plugin.exec(exe, args, cb); };
E.be.plugin.enabled_state = (function(){
    var t = Ci.nsIPluginTag;
    var states = {enabled: t.STATE_ENABLED, disabled: t.STATE_DISABLED,
        clicktoplay: t.STATE_CLICKTOPLAY};
    return function(mime, state, cb){
        if (!(state in states))
            return void cb('invalid state: '+state);
        var tag, phost;
        try {
            phost = Cc["@mozilla.org/plugin/host;1"]
            .getService(Ci.nsIPluginHost);
            if (!phost.getPluginTagForType)
                return void cb('unsupported');
            try { tag = phost.getPluginTagForType(mime); } catch(e){}
            if (!tag)
                return void cb('no plugin for '+mime);
            tag.enabledState = states[state];
            cb();
        } catch(e){ cb(''+e); }
    };
})();

function on_plugin_event(e){ emit(E.be.plugin, 'event', e); }

E.be.mem_info = function(opt, cb){
    var ret;
    try { ret = util.get_mem_info(); }
    catch(ex){ return cb(''+ex); }
    cb(null, ret.total>opt.verbose_above ? ret : _.omit(ret, 'details'));
};
E.be.gen_uuid = function(cb){ cb(null, uuid().number.replace(/[{}-]/g, '')); };

function mv_to_tmpdir(file){
    var tmpdir = fileutil.getDir('TmpD', []);
    var tmpfile = tmpdir.clone();
    tmpfile.append(file.leafName+'.del');
    tmpfile.createUnique(file.NORMAL_FILE_TYPE, 0x0777);
    var mvfile = file.clone();
    mvfile.moveTo(tmpdir, tmpfile.leafName);
}

E.be.download = function(opt, cb){
    try {
        var file = fileutil.File(opt.path);
        if (opt.file)
            file.append(opt.file);
        if (file.exists())
        {
            try { file.remove(false); }
            catch(err){ mv_to_tmpdir(file); }
        }
        var dl = downloader.createInstance(Ci.nsIDownloader);
        dl.init({onDownloadComplete: on_dl_complete.bind(null, opt, cb)},
            file);
        io.newChannel(opt.url, null, null).asyncOpen(dl, file);
    } catch(e){
        console.error('download.init error: '+e);
        cb('dl.init '+e);
    }
};
function on_dl_complete(opt, cb, downloader, req, ctx, status, res){
    var err, ret;
    try {
        if (status)
            return err = 'status: '+util.err_string(status);
        var chan = QI(req, Ci.nsIHttpChannel);
        status = chan.responseStatus;
        if (status!=200)
            return err = status+' '+chan.responseStatusText;
        if (opt.permissions)
            res.permissions = parseInt(opt.permissions, 8);
        ret = {url: chan.URI.asciiSpec, headers: {}};
        chan.visitResponseHeaders({
            visitHeader: function(name, value){ ret.headers[name] = value; }
        });
    } catch(e){ err = {message: e.message, stack: e.stack};
    } finally {
        if (err && res)
        {
            try { res.remove(false); }
            catch(e){ console.error('dl cleanup failed: '+e); }
        }
        cb(err, ret);
    }
}

function calc_hash(hash, path, cb){
    netutil.asyncFetch(fileutil.File(path), function(is, res){
        if (res)
            return void cb(util.err_string(res));
        try {
            var ch = Cc['@mozilla.org/security/hash;1']
            .createInstance(Ci.nsICryptoHash);
            ch.init(ch[hash]);
            var pump = Cc['@mozilla.org/network/input-stream-pump;1']
            .createInstance(Ci.nsIInputStreamPump);
            pump.init(is, -1, -1, 0, 0, true);
            pump.asyncRead({
                onDataAvailable: function(req, ctx, is, offset, count){
                    ch.updateFromStream(is, count); },
                onStopRequest: function(req, ctx, status){
                    if (status)
                        return void cb(util.err_string(status));
                    var buf = ch.finish(false), hash = '';
                    for (var i in buf)
                        hash += ('0'+buf.charCodeAt(i).toString(16)).slice(-2);
                    cb(null, hash);
                },
                onStartRequest: function(){},
            }, null);
        } catch(e){ cb(''+e); }
    });
}
E.be.md5sum = function(path, cb){ calc_hash('MD5', path, cb); };
E.be.sha1sum = function(path, cb){ calc_hash('SHA1', path, cb); };

E.be.file_size = function(path, cb){
    try {
        var file = fileutil.File(path);
        if (!file.exists())
            return void cb(null, -1);
        cb(null, file.fileSize);
    } catch(e){
        console.error('file_size error: '+e);
        cb('file_size '+e);
    }
};

var reg_inited, reg_key;
function reg_init(){
    if (reg_inited)
        return;
    reg_key = registry.createInstance(Ci.nsIWindowsRegKey);
    reg_inited = true;
}
E.be.reg = {};
E.be.reg.init = function(cb){
    try {
        reg_init();
        cb();
    } catch(e){ cb(''+e); }
};
E.be.reg.const = function(cb){
    try {
        reg_init();
        cb(null, _.pick(reg_key, 'ROOT_KEY_CLASSES_ROOT',
            'ROOT_KEY_CURRENT_USER', 'ROOT_KEY_LOCAL_MACHINE', 'TYPE_NONE',
            'TYPE_STRING', 'TYPE_BINARY', 'TYPE_INT', 'TYPE_INT64'));
    } catch(e){ cb(''+e); }
};
E.be.reg.get = function(opt, cb){
    var err, ret;
    try {
        reg_init();
        reg_key.open(opt.root||reg_key.ROOT_KEY_CURRENT_USER, opt.path,
            reg_key.ACCESS_READ);
        if (!reg_key.hasValue(opt.attrib))
            return;
        if (!opt.type)
            opt.type = reg_key.getValueType(opt.attrib);
        switch (opt.type)
        {
        case reg_key.TYPE_STRING:
            ret = reg_key.readStringValue(opt.attrib);
            break;
        case reg_key.TYPE_BINARY:
            ret = reg_key.readBinaryValue(opt.attrib);
            break;
        case reg_key.TYPE_INT:
            ret = reg_key.readIntValue(opt.attrib);
            break;
        case reg_key.TYPE_INT64:
            ret = reg_key.readInt64Value(opt.attrib);
            break;
        default:
            return err = 'type not supported '+opt.type;
        }
    } catch(e){ err = {message: e.message, stack: e.stack};
    } finally {
        reg_key.close();
        cb(err, ret);
    }
};
E.be.reg.set = function(opt, cb){
    var err, ret;
    try {
        reg_init();
        reg_key.open(opt.root||reg_key.ROOT_KEY_CURRENT_USER, opt.path,
            reg_key.ACCESS_WRITE);
        if (!opt.type)
        {
            if (typeof opt.value=='string')
                opt.type = reg_key.TYPE_STRING;
            else if (typeof opt.value=='number')
                opt.type = reg_key.TYPE_INT;
            else
                opt.type = reg_key.TYPE_NONE;
        }
        switch (opt.type)
        {
        case reg_key.TYPE_STRING:
            reg_key.writeStringValue(opt.attrib, opt.value);
            break;
        case reg_key.TYPE_BINARY:
            reg_key.writeBinaryValue(opt.attrib, opt.value);
            break;
        case reg_key.TYPE_INT:
            reg_key.writeIntValue(opt.attrib, opt.value);
            break;
        case reg_key.TYPE_INT64:
            reg_key.writeInt64Value(opt.attrib, opt.value);
            break;
        default:
            return err = 'type not supported '+opt.type;
        }
    } catch(e){ err = {message: e.message, stack: e.stack};
    } finally {
        reg_key.close();
        cb(err, ret);
    }
};
E.be.reg.del = function(opt, cb){
    var err, ret;
    try {
        reg_init();
        reg_key.open(opt.root||reg_key.ROOT_KEY_CURRENT_USER, opt.path,
            reg_key.ACCESS_READ);
        if (!reg_key.hasValue(opt.attrib))
            return;
        reg_key.removeValue(opt.attrib);
    } catch(e){ err = {message: e.message, stack: e.stack};
    } finally {
        reg_key.close();
        cb(err, ret);
    }
};

var v26 = version_util.version_cmp(system.version, '26')>=0 &&
    version_util.version_cmp(system.version, '27')<0;
E.firefox.panel = {};
E.firefox.panel.resize = function(w, h, cb){
    var p = workers.popup && workers.popup.w;
    if (!p || !p.resize || (p.width===w && p.height===h))
        return cb && cb();
    if (conf.debug_msg)
	console.debug('panel resize', w, h);
    p.resize(w, h);
    if (cb)
	cb();
    if (!v26)
        return;
    timers.setTimeout(function(){
        /* XXX bahaa: hack to workaround panel render bug in ff26 */
        if (p)
            p.resize(p.width, p.height+1);
    }, 30);
};
E.firefox.panel.close = function(cb){
    var p = workers.popup && workers.popup.w;
    if (p)
        p.hide();
    cb(p ? null : 'no popup');
};
E.firefox.panel.is_showing = function(cb){
    var p = workers.popup && workers.popup.w;
    cb(null, p && p.isShowing);
};
E.firefox.workers = {};
E.firefox.workers.disconnect = function(id, cb){
    if (!workers[id])
        return cb('no worker with id '+id);
    workers[id].destroy();
    cb();
};

E.management.get = function(id, cb){
    try {
        AddonManager.getAddonByID(id, function(addon){
            cb(null, to_chrome_extension(addon)); });
    } catch(e){ cb(''+e); }
};
E.management.getAll = function(cb){
    try {
        AddonManager.getAllAddons(function(addons){
            cb(null, _.map(addons, to_chrome_extension)); });
    } catch(e){ cb(''+e); }
};

function to_chrome_extension(addon){
    return !addon ? addon : {
        id: addon.id,
        name: addon.name,
        description: addon.description,
        version: addon.version,
        enabled: addon.isActive,
        homepageUrl: addon.homepageURL,
        optionsUrl: addon.optionsURL,
        type: addon.type,
    };
}

E.cookies.get = function(details, cb){
    var url = details.url, name = details.name;
    try {
        var uri = io.newURI(url, null, null);
        var e = cookies.getCookiesFromHost(uri.host);
        while (e && e.hasMoreElements())
        {
            var c = QI(e.getNext(), Ci.nsICookie2);
            if (c && c.host===uri.host && c.name===name)
                return void cb(null, to_chrome_cookie(c));
        }
        cb();
    } catch(ex){ cb(''+ex); }
};

function to_chrome_cookie(cookie){
    return !cookie ? cookie : {
        name: cookie.name,
        value: cookie.value,
        path: cookie.path,
        secure: cookie.isSecure,
        httpOnly: cookie.isHttpOnly,
        session: cookie.isSession,
        expirationDate: cookie.expiry,
    };
}

/* XXX bahaa/arik: add to be_util. a close variation of this exists in
 * pub/be_chrome.js */
function get_obj(obj, sub){
    var a = obj.split('.').concat(sub||[]), o = E;
    for (var i=0; i<a.length && o; i++)
        o = o[a[i]];
    return i<a.length ? null : o;
}

function handle_msg(src, data, s){
    var list = src.list;
    if (conf.debug_msg)
	console.debug('addon got: '+s);
    if (data.msg=='init')
    {
	src.postMessage({cmd: 'init', conf: conf, zon_config: zconf,
	    context: data.context});
        return true;
    }
    var msg = data.msg;
    if (!msg)
	return false;
    var obj = get_obj(msg.obj, msg.sub);
    if (!obj)
        return false;
    var cb;
    switch (data.type)
    {
    case 'be_msg_req':
        if (msg.msg!=='call_api')
            return false;
        if (conf.debug_api)
        {
            console.info(msg.obj+(msg.sub ? '.'+msg.sub : '')+'.'+msg.func+'('+
                JSON.stringify(msg.args)+')');
        }
        var func = obj[msg.func];
        if (!_.isFunction(func))
            return false;
        if (data._tab_id && func===E.be.reload_ext)
        {
            console.log('ignoring reload request from tpopup');
            return true;
        }
        msg.args.push(function(err){
            var args = array.args(arguments).slice(1);
            src.postMessage({id: data.id, has_cb: data.has_cb,
                ret: {last_error: err, args: args}, type: 'be_msg_resp',
                context: data.context});
        });
        func.apply(null, msg.args);
        break;
    case 'be_msg_add_listener':
        if (conf.debug_api)
        {
            console.info(msg.obj+(msg.sub ? '.'+msg.sub : '')+
                '.add_listener()');
        }
        obj.on('event', cb = function(){
            if (conf.debug_api)
            {
                console.log('event for '+msg.obj+'.'+msg.sub+' '+
                    JSON.stringify(array.args(arguments)));
            }
            src.postMessage({id: data.id, has_cb: data.has_cb,
                args: array.args(arguments), type: 'be_msg_event',
                context: data.context});
        });
	cb.__cb_obj = obj;
        list[data.id] = cb;
        break;
    case 'be_msg_del_listener':
        cb = list[data.id];
        if (!cb || cb.__cb_obj!==obj)
	    console.error('obj mismatch for be_msg_del_listener '+data.id);
        obj.removeListener('event', cb);
        delete list[data.id];
        break;
    default:
        console.error('unknown message type: '+s);
        return false;
    }
    return true;
}

function workers_sort(worker){
    switch (worker.name)
    {
    case 'bg': return 2; // bg last
    case 'popup': return 1;
    case 'about': return 0;
    default: return -1; // tpopups first
    }
}

function workers_uninit(){
    _.invoke(_.sortBy(workers, workers_sort), 'destroy');
}

var workers = {};
function Worker(name, w, opt){
    opt = opt||{};
    this.name = name;
    this.list = {};
    this.w = w;
    this.hidden_frame = opt.hidden_frame;
    this.frame = opt.frame;
    this.tpopup = opt.tpopup;
    this.tab_id = opt.tab_id;
    w.on('detach', this.destroy.bind(this));
    workers[name] = this;
}

Worker.prototype.destroy = function(){
    if (this.destroyed)
        return;
    this.destroyed = true;
    console.log(this.name+' uninit');
    if (this.tpopup)
        bg_connection_update(this.name, this.tab_id, false);
    var l = this.list;
    for (var id in l)
    {
        l[id].__cb_obj.removeListener('event', l[id]);
        delete l[id];
    }
    this.list = null;
    if (this.w)
        this.w.destroy();
    this.w = null;
    if (this.hidden_frame)
        hiddenframes.remove(this.hidden_frame);
    this.hidden_frame = null;
    this.frame = null;
    delete workers[this.name];
};

Worker.prototype.postMessage = function(o, s){
    if (this.tpopup)
        o._connection_id = this.name;
    try {
        this.w.postMessage(this.tpopup ? o : s ? s : JSON.stringify(o));
    } catch(e){
        console.error('failed to post message to '+this.name+': '+e); }
};

Worker.prototype.window = function(){
    var e;
    if (this.hidden_frame)
        e = this.hidden_frame && this.hidden_frame.element;
    else
        e = this.frame;
    var win = e && e.contentWindow;
    return (win && win.wrappedJSObject) || win;
};

function bg_connection_update(connection_id, tab_id, connected){
    if (!workers.bg)
        return console.warn('bg_connection_update: no bg');
    workers.bg.postMessage({_connection_id: connection_id, _tab_id: tab_id,
        type: 'be_msg_'+(connected ? 'connect' : 'disconnect')});
}

E.create_background = function(tab){
    if (workers.bg)
	return console.error('bg already exist');
    var opt = {onMessage: on_msg.bind(null, 'bg'),
        contentScriptOptions: {type: 'bg'}, contentScriptWhen: 'start',
        contentScriptFile: data_url('content_script.js')};
    if (tab)
    {
        _.extend(opt.contentScriptOptions, {conf: conf, zon_config: zconf,
            url: conf.url_bext_debug+'/be_bg.html'});
        var w = tab.attach(opt);
        var frame = tabutil.getTabBrowserForTab(to_XUL_tab(tab))
        .contentDocument.getElementById('__bg_iframe');
        new Worker('bg', w, {frame: frame});
        return;
    }
    var hf = hiddenframes.HiddenFrame({
        onUnload: function(){ hiddenframes.remove(hf); },
    });
    hiddenframes.add(hf);
    opt.contentURL = data_url('be_bg.html');
    opt.frame = hf.element;
    new Worker('bg', symbiont.Symbiont(opt), {hidden_frame: hf});
};

function on_msg(name, msg){
    var o = msg;
    if (typeof msg=='string')
        o = JSON.parse(msg);
    else
        msg = JSON.stringify(msg);
    var worker = workers[name];
    if (handle_msg(worker, o, msg))
        return;
    // XXX bahaa: use connection id for popup and about and rm this
    var others = _.compact(name!=='bg' ? [workers.bg] :
        o._connection_id ? [workers[o._connection_id]] :
        [workers.popup, workers.about]);
    if (!others.length)
    {
        if (conf.debug_msg)
            console.warn(name+' ignores:'+msg);
        return;
    }
    _.each(others, function(other){
        if (conf.debug_msg)
            console.debug('passing to '+other.name+': '+msg);
        other.postMessage(o, msg);
    });
}

E.create_popup = function(tab){
    if (workers.popup)
	return console.error('popup already exist');
    var opt = {onMessage: on_msg.bind(null, 'popup'),
        contentScriptOptions: {type: 'popup'}, contentScriptWhen: 'start',
        contentScriptFile: data_url('content_script.js')};
    var p, w;
    if (tab)
    {
        _.extend(opt.contentScriptOptions, {conf: conf, zon_config: zconf,
            url: conf.url_bext_debug+'/be_popup.html'});
	p = new Worker('popup', tab.attach(opt));
        p.w.port.emit('show');
        return;
    }
    _.extend(opt.contentScriptOptions, {url: 'be_popup.html'});
    opt.contentURL = data_url(conf.default_popup);
    p = new Worker('popup', panel(opt));
    w = p.w;
    w.on('show', function(){ w.port.emit('show'); });
    w.on('hide', function(){
        w.port.emit('hide');
        E.firefox.panel.resize(300, 1);
    });
    setup_tooltip_fix(w);
    // XXX bahaa HACK: start with a wide popup to workaround firefox
    // bug not flipping panel when its width is increased dynamically
    E.firefox.panel.resize(300, 1);
    E.browserAction.init({default_icon: 'img/icon19_gray.png',
        default_title: 'Hola Unblocker', panel: w});
};
