// LICENSE_CODE ZON
'use strict'; /*jslint browser:true, es5:true*//*global exports*/
var chrome = require('chrome');
var Ci = chrome.Ci;
var Cc = chrome.Cc;
var Cr = chrome.Cr;
var sdk_self = require('sdk/self');
var storage = require('sdk/simple-storage').storage;
var qs = require('sdk/querystring');
var request = require('sdk/request').Request;
var winutil = require('sdk/window/utils');
var _ = require('./underscore');
var conf = require('./conf').conf;
var mem_mgr = Cc["@mozilla.org/memory-reporter-manager;1"]
.getService(Ci.nsIMemoryReporterManager);
var E = exports;

E.perr = function(id, info, err, filehead){
    var err_msg = !err ? '' : typeof err=='string' ? err : err.message||'';
    info = !info ? err_msg : info+' '+err_msg;
    E._perr({id: id, info: info, bt: err && err.stack,
        filehead: filehead||E.serr()});
};

var perr_sent = {};
E.perr_once = function(id){
    if (perr_sent[id])
        return;
    perr_sent[id] = true;
    E.perr.apply(E, arguments);
};

E._perr = function(opt){
    var uuid = '';
    try { uuid = storage.uuid||''; } catch(e){}
    var id = /^be_/.test(opt.id) ? opt.id : 'be_'+opt.id;
    var info = opt.info||'';
    console.log(id, info);
    request({
        content: {filehead: opt.filehead||'', bt: opt.bt||'',
            info: info, build: opt.build||''},
        url: conf.url_perr+'/perr?'+qs.stringify({
            id: id, uuid: uuid, ver: sdk_self.version, browser: 'firefox'}),
    }).post();
};

E.serr = function(){
    var br = E.gBrowser();
    var navigator = br && br.contentWindow && br.contentWindow.navigator;
    return !navigator ? '' : E.dump_obj(navigator,
        ['userAgent', 'platform', 'product', 'productSub', 'buildID',
        'vendor']);
};

E.dump_obj = function(obj, fields){
    return !obj ? '' :
        E.json(obj, fields, ' ').replace(/^[{ }]\n?|,$/gm, '');
};

E.json = function(o, replacer, space){
    try { return JSON.stringify(o, replacer, space); }
    catch(err){ return '[circular]'; }
};

E.gBrowser = function(){
    var bw = winutil.getMostRecentBrowserWindow();
    return bw && bw.gBrowser;
};

E.dump_interfaces = function(obj){
    var ifs = [];
    for (var i in Ci)
    {
        try {
            obj.QueryInterface(Ci[i]);
            ifs.push(''+i);
        } catch(err){}
    }
    return !ifs.length ? '' : ifs.join(', ');
};

var IMR = Ci.nsIMemoryReporter, mem_result;
var kinds = _.invert({
    non_heap: IMR.KIND_NONHEAP,
    heap: IMR.KIND_HEAP,
    other: IMR.KIND_OTHER,
});

function handle_mem_report(process, path, kind, units, amount, description){
    /* XXX bahaa: rm hardcoded id */
    if (units!==IMR.UNITS_BYTES ||
        !/resource:\\\\jid1-4p0kohsjxu1qgg-at-jetpack/.test(path))
    {
        return;
    }
    var a = path.split('/'), o = mem_result;
    o.total = (o.total||0)+amount;
    var k = kinds[kind];
    o[k] = (o[k]||0)+amount;
    o = o.details;
    for (var i=0; i<a.length; i++)
    {
        o[a[i]] = o[a[i]]||{};
        o = o[a[i]];
    }
    o.bytes = (o.bytes||0)+amount;
}

E.get_mem_info = function(){
    mem_result = {details: {}};
    var f = mem_mgr.enumerateReporters, e = f && f();
    while (e && e.hasMoreElements())
    {
        var r = e.getNext().QueryInterface(IMR);
        handle_mem_report(r.process, r.path, r.kind, r.units, r.amount,
            r.description);
    }
    f = mem_mgr.enumerateMultiReporters;
    e = f && f();
    while (e && e.hasMoreElements())
    {
        var mr = e.getNext().QueryInterface(Ci.nsIMemoryMultiReporter);
        mr.collectReports(handle_mem_report, null);
    }
    var ret = mem_result;
    mem_result = null;
    return ret;
};

E.QI = function(o, i, ignore_err){
    var o2 = null;
    try { o2 = o.QueryInterface(i); }
    catch(e){
        if (!ignore_err)
            console.error('QueryInterface failed:', e);
        /* XXX bahaa: perr per (o, i) */
    }
    return o2;
};

E.err_string = (function(){
    var errors;
    try { errors = _.invert(Cr); }
    catch(err){ errors = {}; }
    errors[Cr.NS_ERROR_PROXY_CONNECTION_REFUSED] =
        'net::ERR_PROXY_CONNECTION_FAILED';
    errors[Cr.NS_ERROR_NET_RESET] = 'net::ERR_CONNECTION_RESET';
    errors[Cr.NS_ERROR_NET_TIMEOUT] = 'net::ERR_TIMED_OUT';
    return function(code){ return errors[code]||'0x'+code.toString(16); };
})();
