// LICENSE_CODE ZON
'use strict'; /*jslint browser:true, es5:true*/ /*global exports*/
var chrome = require('chrome');
var Ci = chrome.Ci;
var Cc = chrome.Cc;
var Cu = chrome.Cu;
var Cr = chrome.Cr;
var prefs = require('sdk/preferences/service');
var unload = require('sdk/system/unload');
var sdk_self = require('sdk/self');
var sysevents = require('sdk/system/events');
var tabutil = require('sdk/tabs/utils');
var xpcom = require('sdk/platform/xpcom');
var Unknown = xpcom.Unknown;
var _ = require('./underscore');
var conf = require('./conf').conf;
var util = require('./ff_util');
var E = exports;
var xutil = Cu.import('resource://gre/modules/XPCOMUtils.jsm', {}).XPCOMUtils;
var netutil = Cu.import('resource://gre/modules/NetUtil.jsm', {}).NetUtil;
var services = Cu.import('resource://gre/modules/Services.jsm', {}).Services;
var ios = services.io;

var defaults, tab_listeners, proxies, proxies_n, callbacks, inited;
var pps = Cc['@mozilla.org/network/protocol-proxy-service;1']
.getService(Ci.nsIProtocolProxyService);

function clear_globals(){
    proxies = {};
    proxies_n = 0;
    defaults = {};
    tab_listeners = {};
    callbacks = {};
}

E.init = function(){
    if (inited)
        return;
    inited = true;
    clear_globals();
    var pac = prefs.get('network.proxy.autoconfig_url', '');
    if (pac.indexOf(conf.pac_cleanup)>=0)
    {
        prefs.reset('network.proxy.autoconfig_url');
        prefs.reset('network.proxy.type');
        pac = prefs.get('network.proxy.autoconfig_url', '');
    }
    defaults.pac = pac;
    defaults.type = prefs.get('network.proxy.type', 5);
    unload.ensure(E, 'uninit');
    start_listening();
};

E.uninit = function(reason){
    if (!inited)
        return;
    inited = false;
    console.log('proxy unload', reason);
    _.each(_.keys(tab_listeners), E.del_tab);
    E.set('');
    stop_listening();
    clear_globals();
};

// order of events for request
// on_opening_request
// ProgressListener.onStateChange with flag STATE_START
// filter.applyFilter: proxy decision is taken here
// on_modify_request: headers and proxy info are available, can modify headers
//                    StreamListener is attached to the request at this stage
// on_examine_response: response headers are available
// StreamListener.onStartRequest: useless
// StreamListener.onDataAvailable: chunk of response data available
// StreamListener.onStopRequest: request finished
function start_listening(){
    sysevents.on('http-on-modify-request', on_modify_request);
    sysevents.on('http-on-opening-request', on_opening_request); // since FF18
    sysevents.on('http-on-examine-response', on_examine_response);
    sysevents.on('http-on-examine-cached-response', on_examine_response);
    sysevents.on('http-on-examine-merged-response', on_examine_response);
    filter.register();
}

function stop_listening(){
    sysevents.off('http-on-modify-request', on_modify_request);
    sysevents.off('http-on-opening-request', on_opening_request);
    sysevents.off('http-on-examine-response', on_examine_response);
    sysevents.off('http-on-examine-cached-response', on_examine_response);
    sysevents.off('http-on-examine-merged-response', on_examine_response);
    filter.unregister();
}

var filter = {registered: false};
filter.applyFilter = function(pps, url, proxy){
    var cb = callbacks.on_find_proxy_for_url;
    log_event('applyFilter', url.asciiSpec);
    if (!cb)
        return proxy;
    var s;
    try { s = cb(url.asciiSpec, url.asciiHost); }
    catch(ex){ console.error('on_find_proxy_for_url error', ex); }
    if (!s || s==='DIRECT')
        return proxy;
    return gen_proxy(s);
};

filter.register = function(){
    if (filter.registered)
        return;
    pps.registerFilter(filter, 0);
    filter.registered = true;
};

filter.unregister = function(){
    if (!filter.registered)
        return;
    pps.unregisterFilter(filter);
    filter.registered = false;
};

function gen_proxy(s){
    if (proxies_n>1000) /* XXX bahaa: improve */
    {
        proxies = {};
        proxies_n = 0;
    }
    if (proxies[s]!==undefined)
        return proxies[s];
    var a = s.split(/;\s*/), p = null, m;
    for (var i=a.length-1; i>=0; i--)
    {
        if (a[i]==='DIRECT')
            continue;
        if (!(m=a[i].match(/PROXY\s*([^:]+)(:(\d+))?/)))
            return void console.error('bad syntax', s, a[i]);
        p = pps.newProxyInfo('http', m[1], m[3]||-1, 0, 0xffffffff, p);
    }
    proxies_n++;
    return proxies[s] = p;
}

E.set = function(url){
    var enable = !!url;
    prefs_set('network.proxy.autoconfig_url', url||defaults.pac);
    prefs_set('network.proxy.type', enable ? 2 : defaults.type);
    console.log('proxy', enable ? 'enabled' : 'disabled');
};

function prefs_set(name, val){
    console.log('prefs.set', name, val);
    prefs.set(name, val);
}

E.add_tab = function(tabid){
    if (tab_listeners[tabid])
        return; /* already listening */
    var tab = get_tab(tabid);
    if (!tab)
        return console.error('no tab with id: '+tabid);
    var br = tabutil.getBrowserForTab(tab);
    if (!br)
        return console.error('no browser for tab id: '+tabid);
    var wp = br.webProgress;
    if (!wp)
        return console.error('no web progress for tab id: '+tabid);
    var t = {};
    t.pl = new ProgressListener(wp, tabid);
    tab_listeners[tabid] = t;
};

E.del_tab = function(tabid){
    var t = tab_listeners[tabid];
    if (!t)
        return;
    if (t.pl)
        t.pl = void t.pl.destroy();
    delete tab_listeners[tabid];
};

function get_tab(id){
    return _.find(tabutil.getTabs(), function(t){
        return tabutil.getTabId(t)===id; });
}

/* per-tab listener used because on-opening-request doesn't always succeed in
 * associating tabid to request. this obviously knows the tabid but it can't
 * replace on-opening-request because it doesn't catch all requests originating
 * from the tab */
function ProgressListener(wp, tabid){
    this.wp = wp;
    this.tabid = tabid;
    wp.addProgressListener(this, wp.NOTIFY_STATE_ALL);
}

var WPL = Ci.nsIWebProgressListener;
ProgressListener.prototype = {
    destroy: function(){
        if (!this.wp)
            return;
        try { this.wp.removeProgressListener(this); }
        catch(ex){ /* tab already removed */ }
        this.wp = null;
    },
    QueryInterface: xutil.generateQI([WPL, Ci.nsISupportsWeakReference]),
    onStateChange: function(wp, req, flags, status){
        if (!(flags & WPL.STATE_START) || !callbacks.on_before_request)
            return;
        log_event('ProgressListener.onStateChange', req);
        req_start(req, this.tabid, wp.DOMWindow.parent===wp.DOMWindow, true);
    },
};

_.each(['on_before_request', 'on_before_send_headers', 'on_headers_received',
    'on_completed', 'on_find_proxy_for_url', 'on_auth_required'],
    function(obj){
    var o = E[obj] = {};
    o.set_listener = function(cb){ callbacks[obj] = cb; };
    o.clear_listener = function(){ delete callbacks[obj]; };
});

function on_modify_request(e){
    var req = e.subject, cb, d, pi, pa;
    if (!(req instanceof Ci.nsIHttpChannel))
        return;
    log_event('on_modify_request', req);
    if (conf.debug_proxy)
    {
        try {
            var s = req.requestMethod+' '+req_url(req)+'\n';
            headers_foreach(req, true, function(name, value){
                s += name+': '+value+'\n'; });
            debug('REQ', s);
        } catch(ex){}
    }
    if ((d=get_req_data(req)))
    {
        d.url = req.URI.asciiSpec; // might change on redirect
        attach_listener(req, d);
        if ((cb=callbacks.on_before_send_headers))
        {
            d.requestHeaders = get_chrome_headers(req, true);
            try { req_op('on_before_send_headers', req, cb(expose_obj(d))); }
            catch(ex){ console.error('on_before_send_headers error', ex); }
            delete d.requestHeaders;
        }
    }
    if (!(req instanceof Ci.nsIProxiedChannel) || !(pi=req.proxyInfo) ||
        pi.type!='http') // includes https (CONNECT)
    {
	return;
    }
    if ((cb=callbacks.on_auth_required))
    {
	try { pa = cb(pi.host); }
	catch(err){ console.error('on_auth_required error', err); }
    }
    else if (pi.host.match(/\.hola\.org$/))
    {
	pa = 'cHJveHk6ZjA3NjFmNzg3NTNk';
        util.perr_once('no_agent_key_ff');
    }
    if (!pa)
	return;
    req.setRequestHeader('Proxy-Authorization', 'Basic '+pa, false);
}

function attach_listener(req, data){
    // we may have already attached a listener if it's a redirected request.
    // also leave the request if no data was attached to it by
    // on_opening_request
    if (!data || data._listening || !callbacks.on_completed ||
        !(req instanceof Ci.nsITraceableChannel))
    {
        return;
    }
    var l = new StreamListener(data);
    var orig = req.setNewListener(l);
    l.set_orig(orig);
    data._listening = true;
}

// get channel's request/response headers formatted like chrome's HttpHeaders
function get_chrome_headers(chan, req){
    var hdrs = [];
    headers_foreach(chan, req, function(name, value){
        hdrs.push(expose_obj({name: name, value: value})); });
    return hdrs;
}

function headers_foreach(chan, req, it){
    chan[req ? 'visitRequestHeaders' : 'visitResponseHeaders'](
        {visitHeader: it});
}

/* stream listener attached to http requests by on_modify_request, to catch
 * request end events. can't rely on on-examine-*-response alone because
 * they're not fired if no response was received from server */
function StreamListener(data){
    this.data = data;
    this.orig = null; /* original stream listener, one we are replacing */
}

StreamListener.prototype = {
    set_orig: function(orig){ this.orig = orig; },
    onStopRequest: function(req, context, code){
        log_event('StreamListener.onStopRequest', req);
        req_end(req, code);
        if (this.orig)
            return this.orig.onStopRequest.apply(this.orig, arguments);
    },
    /* just pass along to original listener */
    onStartRequest: function(req, context){
        log_event('StreamListener.onStartRequest', req);
        if (this.orig)
            return this.orig.onStartRequest.apply(this.orig, arguments);
    },
    onDataAvailable: function(req, context, stream, offset, count){
        log_event('StreamListener.onDataAvailable', req);
        if (this.orig)
            return this.orig.onDataAvailable.apply(this.orig, arguments);
    },
};

var RD_MAGIC = 'RD_'+_.random(0xffffffff);
function set_req_data(req, data){
    if (!(req instanceof Ci.nsIWritablePropertyBag2))
        return console.error('no nsIWritablePropertyBag2 for', req_url(req));
    // can't simply do 'req[prop] = data' because req isn't extensible. so we
    // use the PropertyBag
    var u = new Unknown();
    u.wrappedJSObject = data;
    req.setPropertyAsInterface(RD_MAGIC, u);
}

function get_req_data(req){
    if (!(req instanceof Ci.nsIPropertyBag2))
        return void console.error('no nsIPropertyBag2 for', req_url(req));
    var u = null;
    try { u = req.getPropertyAsInterface(RD_MAGIC, Ci.nsISupports); }
    catch(e){}
    return u && u.wrappedJSObject;
}

function on_opening_request(e){
    if (!callbacks.on_before_request)
        return;
    var req = e.subject;
    log_event('on_opening_request', req);
    var info = get_req_info(req)||{};
    req_start(req, info.id, info.main_frame);
}

/* might get called twice per request by on_opening_request and
 * ProgressListener.onStateChange in this order. first call doesn't always
 * include tabid, the second does but is not always called. */
function req_start(req, id, is_main_frame, from_wpl){
    var cb = callbacks.on_before_request, d, body;
    if (!cb || !req || !(req instanceof Ci.nsIHttpChannel) ||
        req.URI.asciiHost==='local.hola')
    {
        return;
    }
    if (!(d=get_req_data(req)))
        set_req_data(req, d = {requestId: _.uniqueId()});
    d.tabId = id;
    d.url = req.URI.asciiSpec;
    d.is_main = is_main_frame &&
        !!(req.loadFlags & Ci.nsIChannel.LOAD_INITIAL_DOCUMENT_URI);
    d.method = req.requestMethod;
    if (!d.requestBody && (body=get_req_body(req)))
        d.requestBody = expose_obj({str: body});
    try { req_op('on_before_request', req, cb(expose_obj(d))); }
    catch(ex){ console.error('on_before_request error', ex); }
    delete d.requestBody;
}

function req_op(event, req, op){
    var hdrs;
    if (!op)
        return;
    if (!{on_before_request: 1, on_before_send_headers: 1}[event])
        return; // only events before request is sent are supported
    if (op.cancel)
        return void req.cancel(Cr.NS_BINDING_ABORTED);
    if (op.redirectUrl && req.redirectTo) // since FF20
        return void req.redirectTo(ios.newURI(op.redirectUrl, null, null));
    if ((hdrs=op.requestHeaders))
    {
        for (var i=0, l=hdrs.length; i<l; i++)
            req.setRequestHeader(hdrs[i].name, hdrs[i].value, false);
        return;
    }
}

function req_end(req, code){
    var cb = callbacks.on_completed, d;
    if (!cb || !req || !(req instanceof Ci.nsIHttpChannel) ||
        !(d=get_req_data(req)))
    {
        return;
    }
    if (!code)
        fill_req_status(d, req);
    else
        d.error = util.err_string(code);
    try { cb(expose_obj(d)); }
    catch(ex){ console.error('on_completed error', ex); }
}

function on_examine_response(e){
    var req = e.subject, cb, d;
    if (!req || !(req instanceof Ci.nsIHttpChannel))
        return;
    log_event('on_examine_response', req);
    if (conf.debug_proxy)
    {
        try {
            var s = req.requestMethod+' '+req_url(req)+' '
            +req.responseStatus+' '+req.responseStatusText+'\n';
            headers_foreach(req, false, function(name, value){
                s += name+': '+value+'\n'; });
            debug('RESP', s);
        } catch(ex){}
    }
    if (!(cb=callbacks.on_headers_received) || !(d=get_req_data(req)))
        return;
    fill_req_status(d, req);
    d.fromCache = e.type==='http-on-examine-cached-response';
    d.responseHeaders = get_chrome_headers(req, false);
    try { req_op('on_headers_received', req, cb(expose_obj(d))); }
    catch(ex){ console.error('on_headers_received error', ex); }
}

function fill_req_status(d, req){
    var major = 0, minor = 9;
    if (req instanceof Ci.nsIHttpChannelInternal)
    {
        var _major = {}, _minor = {};
        req.getResponseVersion(_major, _minor); // convenient api
        major = _major.value|0;
        minor = _minor.value|0;
    }
    d.statusCode = req.responseStatus;
    d.statusLine = 'HTTP/'+major+'.'+minor+' '+req.responseStatus+' '+
        req.responseStatusText;
}

/* expose all properties for read to unprivileged code */
function expose_obj(obj){
    var p = {};
    delete obj.__exposedProps__;
    _.each(obj, function(val, key){ p[key] = 'r'; });
    obj.__exposedProps__ = p;
    return obj;
}

function log_event(event, req){
    debug('EVENT', event+' '+(typeof req=='string' ? req : req_url(req))); }

var debug_tags = {};
function debug(tag, s){
    if (conf.debug_proxy && debug_tags[tag])
        console.debug(tag+': '+s);
}

function get_req_info(req){
    var info = {};
    if (!(req instanceof Ci.nsIHttpChannel))
        return;
    var ctx = get_context(req.notificationCallbacks) ||
        (req.loadGroup && get_context(req.loadGroup.notificationCallbacks));
    if (!ctx)
        return void debug('TABS', 'no context for '+req.name);
    var win;
    try { win = ctx.associatedWindow; }
    catch(e){ /* might throw NS_ERROR_UNEXPECTED */ }
    if (!win)
        return void debug('TABS', 'no associated window for '+req.name);
    info.main_frame = win.parent===win;
    var br = util.gBrowser();
    if (!br)
    {
        debug('TABS', 'no gBrowser for '+req.name);
        return info;
    }
    try {
        var tabidx = br.getBrowserIndexForDocument(win.top.document);
        if (tabidx<0)
        {
            debug('TABS', 'no tab for '+req.name);
            return info;
        }
        info.id = tabutil.getTabId(br.tabContainer.childNodes[tabidx]);
    }
    catch(e){ debug('TABS', 'no tab for '+req.name); }
    return info;
}

function get_req_body(req){
    var ss, ret, off = -1;
    // XXX bahaa: add PUT and others?
    if (!(req instanceof Ci.nsIHttpChannel) ||
        req.requestMethod!='POST' ||
        !(req instanceof Ci.nsIUploadChannel) ||
        !((ss=req.uploadStream) instanceof Ci.nsISeekableStream))
    {
        return;
    }
    try {
        off = ss.tell();
        ss.seek(Ci.nsISeekableStream.NS_SEEK_SET, 0);
        ret = netutil.readInputStreamToString(ss, ss.available());
        debug('REQ_BODY', ret);
    } catch(ex){ console.error('get_req_body '+req_url(req)+' error: ', ex); }
    finally {
        if (off!=-1)
            try { ss.seek(Ci.nsISeekableStream.NS_SEEK_SET, off); } catch(e){}
    }
    return ret;
}

function get_context(cbs){
    if (!cbs)
        return null;
    try { return cbs.getInterface(Ci.nsILoadContext); }
    catch(err){ return null; }
}

function req_url(req){
    return req instanceof Ci.nsIHttpChannel ? req.URI.asciiSpec : '(unknown)';
}

